import 'package:flutter/material.dart';
import 'package:flutter_riverpod/flutter_riverpod.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/src/provider.dart';

import 'riverpod_endlich_test.mocks.dart';
import 'stuff.dart';

@GenerateMocks([StateListener])
void main() {
  test('can be refreshed', () async {
    var result = TestMachine();
    final container = createContainer();
    final provider = MachineStateProvider<TestMachine, Test>((ref) => result);

    expect(container.read(provider), isFoo());
    expect(container.read(provider.machine), result);
    expect(container.read(provider.machine.notifier), result);

    result = TestMachine(init: (m) => m.toBaz(0));
    expect(container.refresh(provider), isBaz());

    expect(container.read(provider), isBaz());
    expect(container.read(provider.machine), result);
    expect(container.read(provider.machine.notifier), result);
  });

  group('scoping an override overrides all the associated subproviders', () {
    test('when passing the provider itself', () {
      final provider =
          MachineStateProvider<TestMachine, Test>((ref) => TestMachine());
      final root = createContainer();
      final container = createContainer(parent: root, overrides: [provider]);

      expect(container.read(provider.machine).state, isFoo());
      expect(container.read(provider), isFoo());
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine.notifier)
        ]),
      );
      expect(root.getAllProviderElements(), isEmpty);
    });

    test('when using provider.overrideWithValue', () {
      final provider =
          MachineStateProvider<TestMachine, Test>((ref) => TestMachine());
      final root = createContainer();
      final container = createContainer(parent: root, overrides: [
        provider.overrideWithValue(TestMachine(init: (m) => m.toBar(0))),
      ]);

      expect(container.read(provider.machine).state, isBar());
      expect(container.read(provider), isBar());
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine.notifier)
        ]),
      );
      expect(root.getAllProviderElements(), isEmpty);
    });

    test('when using provider.overrideWithProvider', () {
      final provider =
          MachineStateProvider<TestMachine, Test>((ref) => TestMachine());
      final root = createContainer();
      final container = createContainer(parent: root, overrides: [
        provider.overrideWithProvider(
          MachineStateProvider<TestMachine, Test>(
              (ref) => TestMachine(init: (m) => m.toBar(0))),
        ),
      ]);

      expect(container.read(provider.machine).state, isBar());
      expect(container.read(provider), isBar());
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.machine.notifier)
        ]),
      );
      expect(root.getAllProviderElements(), isEmpty);
    });
  });

  // test('overriding listens to the ChangeNotifier', () {}, skip: true);
  // test('overriding family listens to the ChangeNotifier', () {}, skip: true);
  // test('refresh recreates the ChangeNotifier', () {}, skip: true);

  test('family', () {
    final container = createContainer();
    final provider =
        MachineStateProvider.family<TestMachine, Test, int>((ref, value) {
      return TestMachine(init: (m) => m.toFoo(value));
    });

    expect(container.read(provider(0)), isFoo(0));
    expect(container.read(provider(42)), isFoo(42));
  });

  // test('family override', () {
  //   final provider =
  //       ChangeNotifierProvider.family<ValueNotifier<int>, int>((ref, value) {
  //     return ValueNotifier(value);
  //   });
  //   final container = createContainer(overrides: [
  //     provider.overrideWithProvider(
  //       (value) => ChangeNotifierProvider((ref) => ValueNotifier(value * 2)),
  //     ),
  //   ]);

  //   expect(
  //     container.read(provider(0)),
  //     isA<ValueNotifier<int>>().having((source) => source.value, 'value', 0),
  //   );
  //   expect(
  //     container.read(provider(42)),
  //     isA<ValueNotifier<int>>().having((source) => source.value, 'value', 84),
  //   );
  // }, skip: true);

  test('can specify name', () {
    final provider = MachineStateProvider<TestMachine, Test>(
      (ref) => TestMachine(),
      name: 'example',
    );

    expect(provider.name, 'example');

    final provider2 = MachineStateProvider<TestMachine, Test>(
      (ref) => TestMachine(),
    );
    expect(provider2.name, isNull);
  });

  testWidgets('listen to the notifier', (tester) async {
    final notifier = TestMachine();
    final provider = MachineStateProvider<TestMachine, Test>((_) => notifier);

    await tester.pumpWidget(
      ProviderScope(
        child: Consumer(builder: (c, ref, _) {
          return Text(
            ref.watch(provider).i.toString(),
            textDirection: TextDirection.ltr,
          );
        }),
      ),
    );

    expect(find.text('0'), findsOneWidget);
    notifier.toFoo(1);
    await tester.pump();

    expect(find.text('1'), findsOneWidget);

    await tester.pumpWidget(Container());

    expect(notifier.mounted, isFalse);
  });

  test('.notifier obtains the controller without listening to it', () async {
    final dep = StateProvider((ref) => 0);
    final notifier = TestMachine();
    final notifier2 = TestMachine();
    final provider = MachineStateProvider<TestMachine, Test>((ref) {
      return ref.watch(dep).state == 0 ? notifier : notifier2;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(
      provider.machine.notifier,
      (_) => callCount++,
    );

    expect(sub.read(), notifier);
    expect(callCount, 0);

    notifier.toBar(0);

    await container.pump();
    expect(callCount, 0);

    container.read(dep).state++;

    expect(sub.read(), notifier2);

    await container.pump();
    expect(sub.read(), notifier2);
    expect(callCount, 1);
  });

  test(
      'overrideWithValue listens to the notifier, support notifier change, and does not dispose of the notifier',
      () async {
    final provider = MachineStateProvider<TestMachine, Test>((_) {
      return TestMachine();
    });
    final notifier = TestMachine();
    final notifier2 = TestMachine();
    final container = createContainer(overrides: [
      provider.overrideWithValue(notifier),
    ]);
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(provider, (_) => callCount++);
    final notifierSub = container.listen(provider.machine, (_) {});

    expect(sub.read(), isFoo());
    expect(callCount, 0);
    expect(notifierSub.read(), notifier);
    expect(notifier.hasListeners, true);

    notifier.toBar(0);

    await container.pump();
    expect(callCount, 1);

    container.updateOverrides([
      provider.overrideWithValue(notifier2),
    ]);

    await container.pump();
    expect(callCount, 2);
    expect(notifier.hasListeners, false);
    expect(notifier2.hasListeners, true);
    expect(notifier.mounted, true);
    expect(notifierSub.read(), notifier2);

    notifier2.toBar(0);

    await container.pump();
    expect(callCount, 3);

    container.dispose();

    expect(callCount, 3);
    expect(notifier2.hasListeners, false);
    expect(notifier2.mounted, true);
    expect(notifier.mounted, true);
  });

  test('overrideWithProvider preserves the state accross update', () async {
    final provider = MachineStateProvider<TestMachine, Test>((_) {
      return TestMachine();
    });
    final notifier = TestMachine();
    final notifier2 = TestMachine();
    final container = createContainer(overrides: [
      provider.overrideWithProvider(
          MachineStateProvider<TestMachine, Test>((_) => notifier)),
    ]);
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(provider, (_) => callCount++);

    expect(sub.read(), isFoo());
    expect(container.read(provider.machine.notifier), notifier);
    expect(notifier.hasListeners, true);
    expect(callCount, 0);

    notifier.toBar(0);

    await container.pump();
    expect(callCount, 1);

    container.updateOverrides([
      provider.overrideWithProvider(
          MachineStateProvider<TestMachine, Test>((_) => notifier2)),
    ]);

    await container.pump();
    expect(callCount, 1);
    expect(container.read(provider.machine.notifier), notifier);
    expect(notifier2.hasListeners, false);

    notifier.toBaz(0);

    await container.pump();
    expect(callCount, 2);
    expect(container.read(provider.machine.notifier), notifier);
    expect(notifier.mounted, true);

    container.dispose();

    expect(callCount, 2);
    expect(notifier.mounted, false);
  });
}
